<?php
/**
 * SEOmatic plugin for Craft CMS
 *
 * A turnkey SEO implementation for Craft CMS that is comprehensive, powerful,
 * and flexible
 *
 * @link      https://nystudio107.com
 * @copyright Copyright (c) 2019 nystudio107
 */

namespace nystudio107\seomatic\seoelements;

use Craft;
use craft\base\ElementInterface;
use craft\base\Model;
use craft\elements\Category;
use craft\elements\db\ElementQueryInterface;
use craft\events\CategoryGroupEvent;
use craft\events\DefineHtmlEvent;
use craft\gql\interfaces\elements\Category as CategoryInterface;
use craft\models\CategoryGroup;
use craft\models\Site;
use craft\services\Categories;
use Exception;
use nystudio107\seomatic\assetbundles\seomatic\SeomaticAsset;
use nystudio107\seomatic\base\GqlSeoElementInterface;
use nystudio107\seomatic\base\SeoElementInterface;
use nystudio107\seomatic\helpers\ArrayHelper;
use nystudio107\seomatic\helpers\Config as ConfigHelper;
use nystudio107\seomatic\helpers\PluginTemplate;
use nystudio107\seomatic\models\MetaBundle;
use nystudio107\seomatic\Seomatic;
use yii\base\Event;
use yii\base\InvalidConfigException;

/**
 * @author    nystudio107
 * @package   Seomatic
 * @since     3.2.0
 */
class SeoCategory implements SeoElementInterface, GqlSeoElementInterface
{
    // Constants
    // =========================================================================

    public const META_BUNDLE_TYPE = 'categorygroup';
    public const ELEMENT_CLASSES = [
        Category::class,
    ];
    public const REQUIRED_PLUGIN_HANDLE = null;
    public const CONFIG_FILE_PATH = 'categorymeta/Bundle';

    // Public Static Methods
    // =========================================================================

    /**
     * Return the sourceBundleType for that this SeoElement handles
     *
     * @return string
     */
    public static function getMetaBundleType(): string
    {
        return self::META_BUNDLE_TYPE;
    }

    /**
     * Returns an array of the element classes that are handled by this SeoElement
     *
     * @return array
     */
    public static function getElementClasses(): array
    {
        return self::ELEMENT_CLASSES;
    }

    /**
     * Return the refHandle (e.g.: `entry` or `category`) for the SeoElement
     *
     * @return string
     */
    public static function getElementRefHandle(): string
    {
        return Category::refHandle() ?? 'category';
    }

    /**
     * Return the handle to a required plugin for this SeoElement type
     *
     * @return null|string
     */
    public static function getRequiredPluginHandle()
    {
        return self::REQUIRED_PLUGIN_HANDLE;
    }

    /**
     * Install any event handlers for this SeoElement type
     */
    public static function installEventHandlers()
    {
        $request = Craft::$app->getRequest();

        // Install for all requests
        Event::on(
            Categories::class,
            Categories::EVENT_AFTER_SAVE_GROUP,
            function(CategoryGroupEvent $event) {
                Craft::debug(
                    'Categories::EVENT_AFTER_SAVE_GROUP',
                    __METHOD__
                );
                Seomatic::$plugin->metaBundles->resaveMetaBundles(self::META_BUNDLE_TYPE);
            }
        );
        Event::on(
            Categories::class,
            Categories::EVENT_AFTER_DELETE_GROUP,
            function(CategoryGroupEvent $event) {
                Craft::debug(
                    'Categories::EVENT_AFTER_DELETE_GROUP',
                    __METHOD__
                );
                Seomatic::$plugin->metaBundles->resaveMetaBundles(self::META_BUNDLE_TYPE);
            }
        );

        // Install for all non-console requests
        if (!$request->getIsConsoleRequest()) {
            // Handler: Categories::EVENT_AFTER_SAVE_GROUP
            Event::on(
                Categories::class,
                Categories::EVENT_AFTER_SAVE_GROUP,
                function(CategoryGroupEvent $event) {
                    Craft::debug(
                        'Categories::EVENT_AFTER_SAVE_GROUP',
                        __METHOD__
                    );
                    if ($event->categoryGroup !== null && $event->categoryGroup->id !== null) {
                        Seomatic::$plugin->metaBundles->invalidateMetaBundleById(
                            SeoCategory::getMetaBundleType(),
                            $event->categoryGroup->id,
                            $event->isNew
                        );
                        // Create the meta bundles for this category if it's new
                        if ($event->isNew) {
                            SeoCategory::createContentMetaBundle($event->categoryGroup);
                            Seomatic::$plugin->sitemaps->submitSitemapIndex();
                        }
                    }
                }
            );
            // Handler: Categories::EVENT_AFTER_DELETE_GROUP
            Event::on(
                Categories::class,
                Categories::EVENT_AFTER_DELETE_GROUP,
                function(CategoryGroupEvent $event) {
                    Craft::debug(
                        'Categories::EVENT_AFTER_DELETE_GROUP',
                        __METHOD__
                    );
                    if ($event->categoryGroup !== null && $event->categoryGroup->id !== null) {
                        Seomatic::$plugin->metaBundles->invalidateMetaBundleById(
                            SeoCategory::getMetaBundleType(),
                            $event->categoryGroup->id,
                            false
                        );
                        // Delete the meta bundles for this category
                        Seomatic::$plugin->metaBundles->deleteMetaBundleBySourceId(
                            SeoCategory::getMetaBundleType(),
                            $event->categoryGroup->id
                        );
                    }
                }
            );
        }

        // Install only for non-console site requests
        if ($request->getIsSiteRequest() && !$request->getIsConsoleRequest()) {
        }

        // Handler: Category::EVENT_DEFINE_SIDEBAR_HTML
        Event::on(
            Category::class,
            Category::EVENT_DEFINE_SIDEBAR_HTML,
            static function(DefineHtmlEvent $event) {
                Craft::debug(
                    'Category::EVENT_DEFINE_SIDEBAR_HTML',
                    __METHOD__
                );
                $html = '';
                Seomatic::$view->registerAssetBundle(SeomaticAsset::class);
                /** @var Category $category */
                $category = $event->sender ?? null;
                if ($category !== null && $category->uri !== null) {
                    Seomatic::$plugin->metaContainers->previewMetaContainers($category->uri, $category->siteId, true);
                    // Render our preview sidebar template
                    if (Seomatic::$settings->displayPreviewSidebar) {
                        $html .= PluginTemplate::renderPluginTemplate('_sidebars/category-preview.twig');
                    }
                    // Render our analysis sidebar template
// @TODO: This will be added an upcoming 'pro' edition
//                if (Seomatic::$settings->displayAnalysisSidebar) {
//                    $html .= PluginTemplate::renderPluginTemplate('_sidebars/category-analysis.twig');
//                }
                }
                $event->html .= $html;
            }
        );
    }

    /**
     * Return an ElementQuery for the sitemap elements for the given MetaBundle
     *
     * @param MetaBundle $metaBundle
     *
     * @return ElementQueryInterface
     */
    public static function sitemapElementsQuery(MetaBundle $metaBundle): ElementQueryInterface
    {
        $query = Category::find()
            ->group($metaBundle->sourceHandle)
            ->siteId($metaBundle->sourceSiteId)
            ->limit($metaBundle->metaSitemapVars->sitemapLimit);
        if (!empty($metaBundle->metaSitemapVars->structureDepth)) {
            $query->level('<=' . $metaBundle->metaSitemapVars->structureDepth);
        }

        return $query;
    }

    /**
     * Return an ElementInterface for the sitemap alt element for the given MetaBundle
     * and Element ID
     *
     * @param MetaBundle $metaBundle
     * @param int $elementId
     * @param int $siteId
     *
     * @return null|ElementInterface
     */
    public static function sitemapAltElement(
        MetaBundle $metaBundle,
        int        $elementId,
        int        $siteId,
    ) {
        return Category::find()
            ->id($elementId)
            ->siteId($siteId)
            ->limit(1)
            ->one();
    }

    /**
     * Return a preview URI for a given $sourceHandle and $siteId
     * This just returns the first element
     *
     * @param string $sourceHandle
     * @param int|null $siteId
     * @param int|string|null $typeId
     *
     * @return ?string
     */
    public static function previewUri(string $sourceHandle, $siteId, $typeId = null): ?string
    {
        $uri = null;
        $element = Category::find()
            ->group($sourceHandle)
            ->siteId($siteId)
            ->one();
        if ($element) {
            $uri = $element->uri;
        }

        return $uri;
    }

    /**
     * Return an array of FieldLayouts from the $sourceHandle
     *
     * @param string $sourceHandle
     * @param int|string|null $typeId
     *
     * @return array
     */
    public static function fieldLayouts(string $sourceHandle, $typeId = null): array
    {
        $layouts = [];
        $layoutId = null;
        try {
            $categoryGroup = Craft::$app->getCategories()->getGroupByHandle($sourceHandle);
            if ($categoryGroup) {
                $layoutId = $categoryGroup->getFieldLayoutId();
            }
        } catch (Exception $e) {
            $layoutId = null;
        }
        if ($layoutId) {
            $layouts[] = Craft::$app->getFields()->getLayoutById($layoutId);
        }

        return $layouts;
    }

    /**
     * Return the (entry) type menu as a $id => $name associative array
     *
     * @param string $sourceHandle
     *
     * @return array
     */
    public static function typeMenuFromHandle(string $sourceHandle): array
    {
        return [];
    }

    /**
     * Return the source model of the given $sourceId
     *
     * @param int $sourceId
     *
     * @return CategoryGroup|null
     */
    public static function sourceModelFromId(int $sourceId)
    {
        return Craft::$app->getCategories()->getGroupById($sourceId);
    }

    /**
     * Return the source model of the given $sourceId
     *
     * @param string $sourceHandle
     *
     * @return CategoryGroup|null
     */
    public static function sourceModelFromHandle(string $sourceHandle)
    {
        return Craft::$app->getCategories()->getGroupByHandle($sourceHandle);
    }

    /**
     * Return the most recently updated Element from a given source model
     *
     * @param Model $sourceModel
     * @param int $sourceSiteId
     *
     * @return null|ElementInterface
     */
    public static function mostRecentElement(Model $sourceModel, int $sourceSiteId)
    {
        /** @var CategoryGroup $sourceModel */
        return Category::find()
            ->group($sourceModel->handle)
            ->siteId($sourceSiteId)
            ->limit(1)
            ->orderBy(['elements.dateUpdated' => SORT_DESC])
            ->one();
    }

    /**
     * Return the path to the config file directory
     *
     * @return string
     */
    public static function configFilePath(): string
    {
        return self::CONFIG_FILE_PATH;
    }

    /**
     * Return a meta bundle config array for the given $sourceModel
     *
     * @param Model $sourceModel
     *
     * @return array
     */
    public static function metaBundleConfig(Model $sourceModel): array
    {
        /** @var CategoryGroup $sourceModel */
        return ArrayHelper::merge(
            ConfigHelper::getConfigFromFile(self::configFilePath()),
            [
                'sourceId' => $sourceModel->id,
                'sourceName' => (string)$sourceModel->name,
                'sourceHandle' => $sourceModel->handle,
            ]
        );
    }

    /**
     * Return the source id from the $element
     *
     * @param ElementInterface $element
     *
     * @return int|null
     */
    public static function sourceIdFromElement(ElementInterface $element)
    {
        /** @var Category $element */
        return $element->groupId;
    }

    /**
     * Return the (entry) type id from the $element
     *
     * @param ElementInterface $element
     *
     * @return int|null
     */
    public static function typeIdFromElement(ElementInterface $element)
    {
        /** @var Category $element */
        return null;
    }

    /**
     * Return the source handle from the $element
     *
     * @param ElementInterface $element
     *
     * @return string|null
     */
    public static function sourceHandleFromElement(ElementInterface $element)
    {
        $sourceHandle = '';
        /** @var Category $element */
        try {
            $sourceHandle = $element->getGroup()->handle;
        } catch (InvalidConfigException $e) {
        }

        return $sourceHandle;
    }

    /**
     * Create a MetaBundle in the db for each site, from the passed in $sourceModel
     *
     * @param Model $sourceModel
     */
    public static function createContentMetaBundle(Model $sourceModel)
    {
        /** @var CategoryGroup $sourceModel */
        $sites = Craft::$app->getSites()->getAllSites();
        /** @var Site $site */
        foreach ($sites as $site) {
            $seoElement = self::class;
            Seomatic::$plugin->metaBundles->createMetaBundleFromSeoElement($seoElement, $sourceModel, $site->id, null, true);
        }
    }

    /**
     * Create all the MetaBundles in the db for this Seo Element
     */
    public static function createAllContentMetaBundles()
    {
        // Get all of the category groups with URLs
        $categoryGroups = Craft::$app->getCategories()->getAllGroups();
        foreach ($categoryGroups as $categoryGroup) {
            self::createContentMetaBundle($categoryGroup);
        }
    }

    /**
     * @inheritdoc
     */
    public static function getGqlInterfaceTypeName()
    {
        return CategoryInterface::getName();
    }
}
